/*
  Copyright (c) 2007 Gordon Gremme <gremme@zbh.uni-hamburg.de>
  Copyright (c) 2001 Stefan Kurtz <kurtz@zbh.uni-hamburg.de>
  Copyright (c) 2007 Center for Bioinformatics, University of Hamburg

  Permission to use, copy, modify, and distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.

  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <ctype.h>
#include <string.h>
#include "libgtcore/option.h"
#include "libgtcore/versionfunc.h"
#include "libgtcore/warning.h"

#define MAX_LINE_LENGTH  80

static OPrval parse_options(int *parsed_args, int argc, const char **argv,
                            Env *env)
{
  OptionParser *op;
  OPrval oprval;
  op = option_parser_new("[C-file ...]", "Extract Header-File from C-file(s).",
                         env);
  oprval = option_parser_parse(op, parsed_args, argc, argv, versionfunc, env);
  option_parser_delete(op, env);
  return oprval;
}

static char *forbid[] = {
  "static ",
  "typedef ",
  "int main",
  "DECLAREARRAYSTRUCT(",
  "/*@null@*/ static",
  "/*@unused@*/ static"
};

static unsigned char forbiddenstring(Str *line)
{
  size_t slen, i;
  for (i = 0; i < sizeof (forbid) / sizeof (forbid[0]); i++) {
    slen = strlen(forbid[i]);
    if (slen <= (size_t) str_length(line) &&
       !strncmp(forbid[i], str_get(line), slen)) {
      return (unsigned char) 1;
    }
  }
  return 0;
}

static void removecomments(Str *line, int *incomment, Env *env)
{
  unsigned char *buffer;
  unsigned long pos=0, bufpos=0;

  env_error_check(env);

  if (!line || !str_length(line))
    return;

  buffer = env_ma_malloc(env, (size_t) str_length(line) + 1);

  /* remove comments, except for those used for splint: */
  while (pos < str_length(line)) {
    if (*incomment) {
      if (!strncmp(str_get(line) + pos, "*/", (size_t) 2)) {
        *incomment=0;
        pos+=2;
      }
      else
        pos++;
    }
    else {
      if (str_length(line)     >  (unsigned long) 2   &&
          str_get(line)[pos]   == '/' &&
          str_get(line)[pos+1] == '/') {
        break;
      }
      else if (!strncmp(str_get(line) + pos, "/*", (size_t) 2) &&
               (pos + 2 >= str_length(line) || str_get(line)[pos+2] != '@')) {
        *incomment=1;
        pos+=2;
      }
      else
        buffer[bufpos++] = (unsigned char) str_get(line)[pos++];
    }
  }

  /* remove white spaces */
  while (bufpos && buffer[bufpos-1] == ' ')
    bufpos--;
  buffer[bufpos]='\0';

  /* copy back into line */
  memcpy(str_get(line), buffer, (size_t) (bufpos + 1));
  str_set_length(line, bufpos);
  env_ma_free(buffer, env);
}

static void skproto(const char *filename, FILE *fpin, Env *env)
{
  int linenum = 0, startfunction = 1, incomment = 0;
  Str *line;

  env_error_check(env);
  assert(filename && fpin);

  line = str_new(env);

  while (str_read_next_line(line, fpin, env) != EOF) {
    linenum++;
    removecomments(line, &incomment, env);
    if (str_length(line)) {
      if (startfunction) {
        if (isalpha((int) (str_get(line)[0])) ||
            (str_length(line) >= (unsigned long) 3 &&
             strncmp(str_get(line), "/*@", (size_t) 3) == 0)) {
          if (!forbiddenstring(line)) {
            if (str_length(line) >= (unsigned long) MAX_LINE_LENGTH)
              warning("file %s, line %d too long\n", filename, linenum);
            printf("%s", str_get(line));
            if (str_get(line)[str_length(line)-1] == ')') {
              (void) putchar(';');
              (void) putchar('\n');
            }
            else
              startfunction = 0;
            (void) putchar('\n');
          }
        }
      }
      else {
        if (str_length(line) >= (unsigned long) MAX_LINE_LENGTH)
          warning("file %s, line %d too long\n", filename, linenum);
        printf("%s", str_get(line));
        if (str_get(line)[str_length(line)-1] == ')') {
          (void) putchar(';');
          (void) putchar('\n');
          startfunction = 1;
        }
        (void) putchar('\n');
      }
    }
    str_reset(line);
  }

  str_delete(line, env);
}

int gt_skproto(int argc, const char **argv, Env *env)
{
  FILE *fpin;
  int i, parsed_args, had_err = 0;
  env_error_check(env);

  /* option parsing */
  switch (parse_options(&parsed_args, argc, argv, env)) {
    case OPTIONPARSER_OK: break;
    case OPTIONPARSER_ERROR: return -1;
    case OPTIONPARSER_REQUESTS_EXIT: return 0;
  }

  printf("#ifdef __cplusplus\n");
  printf("extern \"C\" {\n");
  printf("#endif\n");

  if (parsed_args == argc)
    skproto("(stdout)", stdin, env);
  else {
    for (i = parsed_args; i < argc; i++) {
      fpin = env_fa_xfopen(env, argv[i], "r");
      skproto(argv[i], fpin, env);
      env_fa_xfclose(fpin, env);
    }
  }

  printf("#ifdef __cplusplus\n");
  printf("}\n");
  printf("#endif\n");

  return had_err;
}
